package de.upb.infexba.ae;

import org.apache.uima.UimaContext;
import org.apache.uima.analysis_component.JCasAnnotator_ImplBase;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.jcas.JCas;
import org.apache.uima.resource.ResourceInitializationException;

import de.upb.infexba.RevenueStatement;
import de.upb.infexba.Sentence;
import de.upb.infexba.Token;


public abstract class InfexBAAnalysisEngine extends JCasAnnotator_ImplBase{

	// -------------------------------------------------------------------
	// PARAMETERS
	// -------------------------------------------------------------------

	/**
	 * Parameter for the activation an deactivation of the analysis engine
	 */
	public static final String PARAM_ACTIVATE = "Activated";
	
	/**
	 * Parameter for the activation and deactivation of the profiling 
	 * functionality of the analysis engine
	 */
	public static final String PARAM_PROFILING = "Profiling";
	
	/**
	 * Indicates whether the analysis engine is activated or deactivated
	 */
	private boolean isActivated = false;
	
	/**
	 * Indicates whether the profiling functionality of the analysis engine 
	 * is activated or deactivated
	 */
	private boolean isProfiling = true;
	

	// -------------------------------------------------------------------
	// INITIALIZATION
	// -------------------------------------------------------------------

	@Override
	public void initialize(UimaContext aContext) throws ResourceInitializationException {
		super.initialize(aContext);
		// Activate or deactivate - default is deactivated
		String activate = (String) aContext.getConfigParameterValue(PARAM_ACTIVATE);
		if ("true".equals(activate))
			isActivated = true;
		else if ("false".equals(activate))
			isActivated = false;		
		// Activate or deactivate profiling - default is deactivated
		String profile = (String) aContext.getConfigParameterValue(PARAM_PROFILING);
		if ("true".equals(profile))
			isProfiling = true;
		else if ("false".equals(profile))
			isProfiling = false;	
		// Profiling
		processedDocuments = 0;
		processedStatements = 0;
		processedSentences = 0;
		processedTokens = 0;
		curMilliseconds = 0;
		totalMilliseconds = 0;
	}
	
	/**
	 * Getter for the activation parameter.
	 * 
	 * @return <CODE>true</CODE> iff. parameter is activated
	 */
	public boolean isActivated(){
		return isActivated;
	}
	
	
	// -------------------------------------------------------------------
	// PROCESSING
	// -------------------------------------------------------------------

	@Override
	public void process(JCas jcas) throws AnalysisEngineProcessException {
		if (isActivated){
			// Profiling
			long start = 0;
			if (isProfiling){
				start = System.currentTimeMillis();	
				processedCandidates += jcas.getAnnotationIndex(RevenueStatement.type).size();
			}
			// Actual processing
			this.processDocument(jcas);
			// Profiling again
			long end = 0; 
			if (isProfiling) {
				end = System.currentTimeMillis();
				curMilliseconds = end-start;
				totalMilliseconds += curMilliseconds;
				processedTokens += jcas.getAnnotationIndex(Token.type).size();
				processedSentences += jcas.getAnnotationIndex(Sentence.type).size();
				processedStatements += jcas.getAnnotationIndex(RevenueStatement.type).size();
				processedDocuments += 1;
			}
		}	
	}
	
	/**
	 * Abstract method that is called by the process method. Every processing of
	 * a document shall take place here!
	 * 
	 * @param jcas The JCas object
	 */
	protected abstract void processDocument(JCas jcas);

	
	// -------------------------------------------------------------------
	// PROFILING
	// -------------------------------------------------------------------

	/**
	 * The number of processed tokens
	 */
	private int processedTokens;
	
	/**
	 * The number of processed sentences
	 */
	private int processedSentences;

	/**
	 * The number of processed candidate statements
	 */
	private int processedCandidates;
	
	
	/**
	 * The number of processed statements
	 */
	private int processedStatements;
	
	/**
	 * The number of documents processed by the analysis engine
	 */
	private int processedDocuments;
	
	/**
	 * The number of milliseconds that was needed for the last processing
	 * of a document
	 */
	private long curMilliseconds;
	
	/**
	 * The total number of milliseconds that was needed for all processed
	 * documents
	 */
	private long totalMilliseconds;
	

	
	// -------------------------------------------------------------------
	// DESTROY
	// -------------------------------------------------------------------


	@Override
	public void destroy() {
		super.destroy();
		if (isActivated && isProfiling) this.printProfilingResults();
	}
	
	/**
	 * Print the results of profiling
	 */
	private void printProfilingResults(){
		double total = totalMilliseconds/1000.0;
		double perDoc = 0;
		if (processedDocuments > 0)
			perDoc = Math.round(10000*(totalMilliseconds/(1000.0*processedDocuments)))/10000.0;
		double perCandidate= 0.0;
		if (processedCandidates > 0) 
			perCandidate = Math.round(10000*(totalMilliseconds/(1000.0*processedCandidates)))/10000.0;
		double perStatement = 0.0;
		if (processedStatements > 0) 
			perStatement = Math.round(10000*(totalMilliseconds/(1000.0*processedStatements)))/10000.0;
		double perSent = 0;
		if (processedSentences > 0)
			perSent = Math.round(10000000*(totalMilliseconds/(1000.0*processedSentences)))/10000.0;
		double perToken = 0;
		if (processedTokens > 0)
			perToken = Math.round(10000000*(totalMilliseconds/(1000.0*processedTokens)))/10000.0;
		System.out.println(
				this.getClass().getSimpleName() + 
				this.getSpace(this.getClass().getSimpleName()) + 
				total + "s/Coll\t" + 
				perDoc+ "s/Doc\t" + 
				perCandidate + "s/Cand\t" + 
				perStatement + "s/Stmt\t" + 
				perSent + "s/1000 Sent\t" + 
				perToken + "s/1000 Tok");
	}
	
	/**
	 * Return the length of white space to be filled in after the given 
	 * word when printing it
	 * @param word The word
	 * @return The white space as a string
	 */
	private String getSpace(String word){
		String space = "";
		for (int i=1; i<= 33-word.length(); i++) space = space + " ";
		return space;
	}

}
