package de.upb.infexba.ae.entity;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.uima.UimaContext;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;
import org.apache.uima.resource.ResourceInitializationException;

import de.upb.infexba.MonetaryExpression;
import de.upb.infexba.TimeExpression;
import de.upb.infexba.ae.InfexBAAnalysisEngine;
import de.upb.infexba.linguistic.MonetarySubexpression;
import de.upb.infexba.linguistic.RegexMoney;
import de.upb.infexba.linguistic.RegexTime;

/**
 * <p>
 * This analysis engine represents a pattern matcher for finding and 
 * automatically annotating monetary expressions in a German text.
 * <p>
 * The pattern matcher uses the regular expressions given in 
 * {@link RegexMoney} to find all monetary subexpressions.
 * <p>
 * Based on these subexpressions, the complete monetary expressions are
 * determined by finding sequences of subexpressions that fulfill certain
 * criteria. Currently, the following rules are used:
 * <ul>
 *   <li><b>IF</b> there are more than two characters between <code>subexpressions(cur)</code>
 *       and <code>subexpressions(cur-1)</code>
 *       <b>THEN</b> <code>subexpressions(cur)</code> is the start of the next monetary 
 *       expression.</li>
 *   <li>...</li>
 * </ul>
 * 
 * @see {@link MonetarySubexpression} 
 * @author hwachsmuth
 *
 */
public class TimeExpMatcher extends InfexBAAnalysisEngine {

	/**
	 * The Regex pattern for time expressions
	 */
	Pattern pattern;
	
	@Override
	public void initialize(UimaContext aContext) throws ResourceInitializationException {
		super.initialize(aContext);
		try {
			// Compile patterns
			pattern = Pattern.compile(RegexTime.TIME_EXPRESSION);
		}
		catch (Exception e){
			e.printStackTrace();
		}
	}
	
	@Override
	public void processDocument(JCas jcas){
		findExpressions(jcas);
	}
	
	/**
	 * Computes and creates all time expressions found in the text that is
	 * contained in the <code>jcas</code> object matched against the given 
	 * <code>regex</code>.
	 * 
	 * @param jcas The JCas object containing the text to be analyzed
	 * @param regex The regular expression to be matched against the text
	 */
	private void findExpressions(JCas jcas){
		String textToMatch = jcas.getDocumentText();
		Matcher matcher = pattern.matcher(textToMatch);
		
		int index = 0;
		while (matcher.find(index)){
			
			// Check if money information intersects with the found string
			MonetaryExpression moneyExp = null;
			Annotation time = new Annotation(jcas, matcher.start(), matcher.end());
			FSIterator<Annotation> moneyIter = jcas.getAnnotationIndex(MonetaryExpression.type).iterator();	
			while(moneyIter.hasNext()){
				MonetaryExpression money = (MonetaryExpression) moneyIter.next();
				 if ((money.getEnd() >= time.getBegin() && money.getEnd() < time.getEnd())
						 |(money.getBegin() >= time.getBegin() && money.getBegin() < time.getEnd())){
					 if (money.getBegin() == 0 || jcas.getDocumentText().charAt(money.getBegin()-1) != '(')
						 moneyExp = money;
				 }
			}
			// Intersection with money information
			if (moneyExp != null) index = moneyExp.getEnd();
			
			// No intersection with money information
			else {
				Annotation timeExpression = new TimeExpression(jcas, matcher.start(), matcher.end());
				if (!noSpaceFollows(textToMatch, matcher.end()))
					timeExpression.addToIndexes();
				index = timeExpression.getEnd();
			}
		};
	}
	
	private boolean noSpaceFollows(String text, int pos){
		if (pos >= text.length()-2) return false;
		return (text.substring(pos, pos+1).matches("[\\d\\w]")
				|| text.substring(pos, pos+2).matches("\\.[\\d\\w]"));
	}
}
