package de.upb.infexba.ae.preprocessing;

import org.apache.uima.cas.FSIterator;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;

import de.upb.infexba.Sentence;
import de.upb.infexba.Token;
import de.upb.infexba.ae.InfexBAAnalysisEngine;
import de.upb.infexba.util.CharTools;
import de.upb.infexba.util.WordLists;

/**
 * 
 * <p>
 * This analysis engine represents a basic rule-based tokenizer 
 * which, in general, can be used to split any German text into token. 
 * However, the rules used by the splitter are specified with German news 
 * articles given as plain text in mind. 
 * <p>
 * Namely, the essential currently used rules for classyfing whether to 
 * split the tokens at the current text position are:
 * <ol>
 *   <li>Always split before a whitespace character.</li>
 * 
 *   <li>Split after a letter if a special character follows which is 
 *     not a period or a hyphen and does not belong to a URL.</li>
 *   <li>Split after a letter if a hyphen follows and if the letter is
 *     preceded by a whitespace.</li>
 *   <li>Split after a letter if a period follows and if the letter is
 *     preceded by a whitespace or the period belongs to an abbreviation.</li>    
 *  
 *   <li>Split after a comma if a letter follows.</li>
 * 
 *   <li>Split after a hyphen if it a term with more than one character precedes.</li>
 * 
 *   <li>Split after a non-comma special character if it does not belong to a URL.</li>
 * </ol>
 * <p>
 * This analysis engine is dependent from a sentence splitter, in that it needs
 * sentence annotations as input to work.
 * <p>
 * There are no results on the accuracy of this tokenizer. Please
 * evaluate it!
 * 
 * @author hwachsmuth
 *
 */
public class RBTokenizer extends InfexBAAnalysisEngine{

	/**
	 * Word list used for recognizing URLs
	 */
	protected static String URL_BEGINS = WordLists.loadWordList(WordLists.URL_BEGINS);
		
	@Override
	public void processDocument(JCas jcas){
		tokenizeDocument(jcas);
	}
	
	/**
	 * Splits the document text given in the <code>JCas</code> object into its
	 * contained tokens. That is, it creates a new annotation for every
	 * found token and adds it to the indices. Whitespaces are skipped.
	 * <p>
	 * This method is sentence-based, i.e., it tries to find all tokens per sentence.
	 * 
	 * @param jcas The JCas object
	 */
	private void tokenizeDocument(JCas jcas){
		FSIterator<Annotation> sentenceIter = jcas.getAnnotationIndex(Sentence.type).iterator();
		while (sentenceIter.hasNext()){
			Sentence sentence = (Sentence) sentenceIter.next();
			this.tokenizeSentence(jcas, sentence);
		}
	}
	
	/**
	 * Finds and creates tokens in the given sentence.
	 * 
	 * @param jcas The JCas object
	 * @param sentence The sentence
	 */
	private void tokenizeSentence(JCas jcas, Sentence sentence){
		int tokenBegin = sentence.getBegin();
		// Skip maybe possible whitespace at beginning of sentence
		while (tokenBegin < sentence.getEnd() 
				&& Character.isWhitespace(jcas.getDocumentText().charAt(tokenBegin)))
			tokenBegin++;
		while (tokenBegin < sentence.getEnd()){
			// Compute next token
			int tokenEnd = getEndOfToken(jcas, sentence, tokenBegin);
			Annotation token = createToken(jcas, tokenBegin, tokenEnd);
			token.addToIndexes();
			// Skip all whitespace characters between two tokens
			tokenBegin = token.getEnd();
			while (tokenBegin < sentence.getEnd() 
					&& Character.isWhitespace(jcas.getDocumentText().charAt(tokenBegin)))
				tokenBegin++;
		}
	}
	
	/**
	 * Determines the end index of the token starting at the given 
	 * index <code>begin</code>. This method based on several heuristic rules. 
	 * Notice that most rules implictly use the notion that a token just ended
	 * beforehand.
	 *  
	 * @param jcas The JCas object
	 * @param sentence The sentence to find the next token in
	 * @param begin The start index of the sentence to be found
	 * @return The end index of the current token
	 */
	private int getEndOfToken(JCas jcas, Sentence sentence, int begin){
		String text = jcas.getDocumentText();
		boolean isURL = false;
		
		// Search for end index of current token
		for (int end = begin; end<sentence.getEnd(); end++){
			// Set characters
			char cur = text.charAt(end); 
			char succ = ' ';
			if (end<sentence.getEnd()-1) succ = text.charAt(end+1);
			char succ2 = ' ';
			if (end<sentence.getEnd()-2) succ2 = text.charAt(end+2);
			char prev = ' ';
			if (end>0) prev = text.charAt(end-1);
			

			// Always split if whitespace follows or if end of sentence
			if (Character.isWhitespace(succ) || end == sentence.getEnd()-1) 
				return end+1;
			
			// URL has special rule
			if (!Character.isSpaceChar(succ2) && text.substring(begin, end+2).matches(URL_BEGINS))
				isURL = true;
			
			// Maybe Split after letter if special character follows and token does not refer to a URL
			if (Character.isLetter(cur) && !Character.isLetterOrDigit(succ) && !isURL){
				// Split if following character is neither period nor hyphen
				if (!CharTools.isHyphen(succ) && !CharTools.isPeriod(succ))
					return end+1;
				// Split if hyphen follows and letter belongs to a word
				else if (CharTools.isHyphen(succ) && !Character.isWhitespace(prev))
					return end+1;
				// Split if period follows and letter belongs to a non-abbrev. word
				else if (CharTools.isPeriod(succ) && !Character.isWhitespace(prev) &&
						!periodBelongsToAnAcronym(text, sentence, end+1)){
					return end+1;
				}
			}
			
			// Split after digit...
			else if (Character.isDigit(cur)){
				// ... if comma follows but does not belong to number
				if (CharTools.isComma(succ) && !Character.isDigit(succ2)) 
					return end+1;
				// ... or if period follows that does not refer to an ordinal number
				if (CharTools.isPeriod(succ) && sentence.getEnd() == end+2)
					return end+1;
				// ... or if digit is followed by a special character that does not belong to a url
				else if (!Character.isLetterOrDigit(succ) && !CharTools.isComma(succ) && !CharTools.isPeriod(succ) && !isURL)
					return end+1;
			}
			
			// Split after hyphen if it does not connect a single character to the following text		
			else if (CharTools.isHyphen(cur) && !isURL){
				if (begin+1 != end) return end+1;
			}
			
			// Split after period if special character follows
			else if (CharTools.isPeriod(cur) && !Character.isLetterOrDigit(succ) && !isURL)
				return end+1;
			
			
			// Mail address is interpreted like a URL
			else if (CharTools.isAt(succ)) isURL = true;
			
			// Split after a non-comma special character if it does not belong to a URL
			else if (!Character.isLetterOrDigit(cur) && !CharTools.isComma(cur) && !CharTools.isPeriod(cur) && !isURL)
				return end+1;
		}
		return sentence.getEnd();
	}
	
	
	/**
	 * Creates and returns a token annotation.
	 * @param jcas The JCas object
	 * @param begin The begin index of the token
	 * @param end The end index of the token
	 * @return The token
	 */
	private Annotation createToken(JCas jcas, int begin, int end){
//		switchAnnot = !switchAnnot;
//		if (switchAnnot) 
			return new Token(jcas, begin, end);
//		else return new Category(jcas, begin, end);
	}

//	private boolean switchAnnot = false;
	
	/**
	 * Determines wheter the period at index <code>i</code> belongs to an acronym
	 * that is on the used acronym list.
	 * 
	 * @param text The text to be analysed
	 * @param i The index of the period
	 * @return <code>true</code> iff. period belongs to an acronym
	 */
	private boolean periodBelongsToAnAcronym(String text, Sentence sentence, int i){
		while (i < sentence.getEnd() && 
				(Character.isLetter(text.charAt(i)) || CharTools.isPeriod(text.charAt(i))))
			i++;
		int begin = i-1;
		while(begin > 0 
				&& (Character.isLetter(text.charAt(begin)) 
						|| CharTools.isPeriod(text.charAt(begin))))
			begin--;
		return WordLists.isAcronym(text.substring(begin+1,i));	
	}

}
