/*
 * Copyright (C) 2008 www.webis.de
 */
package de.upb.infexba.ae.preprocessing;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import org.apache.log4j.Logger;


/**
 * Wraps the German TreeTagger.
 * 
 * 
 * @author peter.prettenhofer@uni-weimar.de
 * @version aitools 2.0
 * 
 *          Created on Jun 18, 2010
 * 
 *          $Id$
 */
public class TreeTaggerWrapper
{



    /**
     * 
     * A class representing a tagged token.
     * 
     * Encapsulates the token, its part-of-speech, and its lemma.
     * 
     * @author peter.prettenhofer@uni-weimar.de
     * @version aitools 2.0
     * 
     *          Created on Jun 18, 2010
     * 
     *          $Id$
     */
    static public class TaggedToken
    {
        static public final TaggedToken NULL  = new TaggedToken("", "", "");

        String                          term  = "";
        String                          pos   = "";
        String                          lemma = "";
        String                          chunk = "";



        public TaggedToken(String term, String pos, String lemma)
        {
            this.term = term;
            this.pos = pos;
            this.lemma = lemma;
        }



        public TaggedToken(String term, String pos, String lemma, String chunk)
        {
            this(term, pos, lemma);
            this.chunk = chunk;
        }



        public TaggedToken(String[] fields)
        {
            if (fields.length < 3) {
                throw new IllegalArgumentException("fields.length must be at least 3.");
            }
            this.term = fields[0];
            this.pos = fields[1];
            this.lemma = fields[2];
            if (fields.length > 3)
                this.chunk = fields[3];
        }



        public String toString()
        {
            return String.format("%s %s %s %s", term, lemma, pos, chunk);
        }
    }



    /**
     * 
     * Wraps the TreeTagger process and provides convenient access to input/output pipes.
     * 
     * 
     * @author peter.prettenhofer@uni-weimar.de
     * @version aitools 2.0
     * 
     *          Created on Jun 18, 2010
     * 
     *          $Id$
     */
    class ProcessWrapper
    {
        Process               p;
        public PrintWriter    in;
        public BufferedReader out;
        public BufferedReader err;
        private boolean closed;

        public ProcessWrapper(String... args) throws IOException
        {
//            System.out.println(Arrays.toString(args));
            ProcessBuilder pb = new ProcessBuilder(args);
            pb.directory(new File(treeTaggerDirectory));
            p = pb.start();
            out = new BufferedReader(new InputStreamReader(p.getInputStream()));
            err = new BufferedReader(new InputStreamReader(p.getErrorStream()));
            in = new PrintWriter(new BufferedWriter(new OutputStreamWriter(p.getOutputStream())));
            try {
                Thread.sleep(2000);
            }
            catch (InterruptedException e) {
                System.out.println(e);
            }

            // Suppressing startup info to stdout
            while (out.ready()) {
//                System.err.println(out.readLine());
            	out.readLine();
            }

            // Suppressing startup info to stderr
            while (err.ready()) {
//                System.err.println(err.readLine());
            	err.readLine();
            }
            closed = false;
        }

        public void close()
        {
            
            try {
                in.close();
                out.close();
                closed = true;
            }
            catch (IOException e) {
                System.err.println("Cannot close input stream.");
                System.err.println(e.getMessage());
            }
        }
        
        public boolean isClosed() {
            return closed;
        }

    }



    class StdErrReader implements Runnable
    {
        private ProcessWrapper process;

        public StdErrReader(ProcessWrapper process) {
            this.process = process;
        }
        
        @Override
        public void run()
        {
            try {
                String line;
                while (true) {
                	line = process.err.readLine();
                    if (line != null) {
//                        System.err.println(line);
                    }
                    if (process.isClosed()) {
                        break;
                    }
                }
            }
            catch (IOException e) {
                throw new RuntimeException(e);
            }

        }
    }

    protected static final String   TAGGER_INPUT_SENT_DELIM   = ".";
    protected static final String[] TAGGER_OUTPUT_SENT_DELIM  = new String[] { ".", "$.", "." };
    protected static final int      NUM_RESIDUE               = 3;
    protected static final String   CHUNKER_INPUT_SENT_DELIM  = ".-$.";
    protected static final String[] CHUNKER_OUTPUT_SENT_DELIM = new String[] { ".-$.", "$./0" };
    
    private static final Logger LOGGER = Logger.getLogger(TreeTaggerWrapper.class);

    private String                  treeTaggerDirectory;

    private ProcessWrapper          taggerProcess;
    private ProcessWrapper          chunkerProcess;

    static int numInstances = 0;
    int instanceId = 0;

    public TreeTaggerWrapper(String treeTaggerDirectory)
    {
        instanceId = ++numInstances;
        LOGGER.info(String.format("Create TreeTaggerWrapper object nr. %d\n", instanceId));
        this.treeTaggerDirectory = treeTaggerDirectory;
        try {
            taggerProcess = new ProcessWrapper(treeTaggerDirectory + "/bin/tree-tagger", "-token", "-lemma",
                "-no-unknown", "-pt-with-lemma", "lib/german.par");
        }
        catch (IOException e) {
            LOGGER.error("Cannot create TreeTagger process - check tree tagger directory and binary.");
            throw new RuntimeException("Cannot create TreeTagger process - check tree tagger directory and binary.");
        }

        try {
            chunkerProcess = new ProcessWrapper(treeTaggerDirectory + "/bin/tree-tagger", "-token", "-eps",
                "0.00000001", "-hyphen-heuristics", "-quiet", "lib/german-chunker.par");
        }
        catch (IOException e) {
            LOGGER.error("Cannot create TreeTagger process - check tree tagger directory and binary.");
            throw new RuntimeException("Cannot create TreeTagger process - check tree tagger directory and binary.");
        }

        Thread taggerStderrReader = new Thread(new StdErrReader(taggerProcess));
        Thread chunkerStderrReader = new Thread(new StdErrReader(chunkerProcess));
        taggerStderrReader.start();
        chunkerStderrReader.start();
    }



    /**
     * Tags the given token sequence and produces a {@link TaggedToken} sequence.
     * 
     * Note: For each call to this method, a new process has to be forked. So make sure that you accumulate token
     * sequences (e.g., documents) and call this method infrequently rather than calling it with small token sequences
     * (e.g., sentences).
     * 
     * @param tokenSeq
     *            The input sequence.
     * @return A sequence of {@link TaggedToken} of equal length.
     */
    public TaggedToken[] tag(final String[] tokenSeq)
    {
        LOGGER.debug(String.format("Tagging %d tokens.\n", tokenSeq.length));
        final List<TaggedToken> taggedSeq = new LinkedList<TaggedToken>();
        final int numTokens = tokenSeq.length;

        try {

            Thread writer = new Thread(new Runnable()
            {

                @Override
                public void run()
                {
                    int counter = 0;
                    // write to stdin of TreeTagger
                    for (String token : tokenSeq) {
                        taggerProcess.in.println(token);
                        counter += 1;
                    }
                    for (int i = 0; i < NUM_RESIDUE; i++) {
                        taggerProcess.in.println(TAGGER_INPUT_SENT_DELIM);
                    }
                    taggerProcess.in.flush();
//                    System.err.printf("%d tokens written to TT. \n", counter);
                }

            });

            Thread stdoutReader = new Thread(new Runnable()
            {

                @Override
                public void run()
                {
                    try {
                        //                        System.err.println("stdout reader running...");
                        // read from stdout of TreeTagger
                        String line;
                        // whether there are residual SENT_DELIM in the pipe
                        boolean residua = true;
                        int tokenCount = 0;
                        TaggedToken token;
                        while (true) {
                            if ((line = taggerProcess.out.readLine()) != null) {
                                String[] fields = line.split("\\s+");
                                if (fields.length < 3) {
                                    continue;
                                }
                                else {
                                    if (residua && !Arrays.equals(fields, TAGGER_OUTPUT_SENT_DELIM)) {
                                        residua = false;
                                    }
                                    if (!residua) {
                                        token = new TaggedToken(fields);
                                        taggedSeq.add(token);
                                        tokenCount += 1;
                                    }
                                }
                                //System.out.printf("%s\t%d\t%d\n", line, tokenCount, numTokens);
                            }
                            if (tokenCount >= numTokens) {
                                break;
                            }
                        }
//                        System.err.printf("%d tokens read from TT. \n", tokenCount);
                    }
                    catch (IOException e) {
                        LOGGER.error(e);
                        throw new RuntimeException(e);
                    }
                }
            });



            writer.start();
            stdoutReader.start();

            try {
                writer.join();
                stdoutReader.join();
            }
            catch (InterruptedException e) {
                throw new RuntimeException(e);
            }
            return taggedSeq.toArray(new TaggedToken[0]);
        }
        catch (RuntimeException e) {
            LOGGER.error(e);
            throw new RuntimeException(e);
        }

    }



    public TaggedToken[] chunk(String[] tokenSeq)
    {
        TaggedToken[] taggedSeq = tag(tokenSeq);
        return chunk(taggedSeq);
    }



    public TaggedToken[] chunk(final TaggedToken[] tokenSeq)
    {
        final List<TaggedToken> taggedSeq = new LinkedList<TaggedToken>();
        final int numTokens = tokenSeq.length;

        try {

            Thread writer = new Thread(new Runnable()
            {

                @Override
                public void run()
                {
                    int count = 0;
                    // write to stdin of TreeTagger
                    for (TaggedToken token : tokenSeq) {
                        chunkerProcess.in.println(String.format("%s-%s", token.term, token.pos));
                        count += 1;
                    }
                    for (int i = 0; i < NUM_RESIDUE; i++) {
                        chunkerProcess.in.println(CHUNKER_INPUT_SENT_DELIM);
                    }
//                    System.err.printf("%d tokens written to TC.\n", count);
                    chunkerProcess.in.flush();
                }
            });

            Thread stdoutReader = new Thread(new Runnable()
            {

                @Override
                public void run()
                {
                    try {
                        String line;
                        int i = 0;
                        boolean residua = true;
                        while (true) {
                            if ((line = chunkerProcess.out.readLine()) != null) {
                                String[] fields = line.split("\\s+");
                                if (residua && !Arrays.equals(fields, CHUNKER_OUTPUT_SENT_DELIM)) {
                                    residua = false;
                                }
                                if (residua) {
                                    continue;
                                }

                                String term = fields[0].substring(0, fields[0].lastIndexOf("-"));
                                TaggedToken token = tokenSeq[i];
                                if (!term.equals(token.term)) {
                                    LOGGER.error(String.format("%s\t%s\n", term, token.term));
                                }
                                assert term.equals(token.term);
                                String chunk = fields[1].substring(fields[1].lastIndexOf("/") + 1);
                                taggedSeq.add(new TaggedToken(token.term, token.pos, token.lemma, chunk));
                                i += 1;
                            }
                            if (i >= numTokens) {
                                break;
                            }
                        }
                    }
                    catch (IOException e) {
                        LOGGER.error(e);
                        throw new RuntimeException(e);
                    }

                }
            });

            writer.start();
            stdoutReader.start();
            try {
                writer.join();
                stdoutReader.join();
            }
            catch (InterruptedException e) {
                throw new RuntimeException(e);
            }

            assert tokenSeq.length == taggedSeq.size();
            return taggedSeq.toArray(new TaggedToken[0]);
        }
        catch (RuntimeException e) {
            LOGGER.error(e);
            throw new RuntimeException(e);
        }
    }



    public void close()
    {
        LOGGER.info(String.format("Closing TreeTaggerWrapper instance nr.: %d", instanceId));
        chunkerProcess.close();
        taggerProcess.close();
    }



    /**
     * @param args
     * @throws IOException
     */
    public static void main(String[] args) throws IOException
    {
        TreeTaggerWrapper tagger = new TreeTaggerWrapper(args[0]);
        try {
        //String[] tokenSeq = "Das ist ein Test .".split(" ");
        String[] tokenSeq = { "Das", "ist", "ein", "Test", "." };
        String[] tokenSeqB = { "Den", "Inhalt", "vor", "Feuchtigkeit", "schützen", "." };

        StringBuilder content = new StringBuilder();
        BufferedReader in = new BufferedReader(new FileReader("/home/peter/share/macbeth.txt"));
        for (String line = ""; (line = in.readLine()) != null;) {
            content.append(line);
        }
        StringTokenizer st = new StringTokenizer(content.toString());
        LinkedList<String> tokens = new LinkedList<String>();
        int i = 0;
        while (st.hasMoreTokens()) {
            String token = st.nextToken();
            token = token.trim();
            if (token == "" || token == "\n")
                break;
            tokens.add(token);
            i += 1;
            if (i > 20000) {
                break;
            }
        }
        tokenSeq = tokens.toArray(new String[0]);
        long t0 = System.currentTimeMillis();
        TaggedToken[] taggedSeq = tagger.chunk(tokenSeq);
        System.out.printf("Took %d msec.", System.currentTimeMillis() - t0);
        System.out.println("taggedSeq.length=" + taggedSeq.length);
        System.out.println("tokenSeq.length=" + tokenSeq.length);
        for (int j = 16000; j < taggedSeq.length; j++) {
            System.out.printf("%s - %s\n", tokenSeq[j], taggedSeq[j]);
        }
        taggedSeq = tagger.chunk(tokenSeqB);
        for (TaggedToken token : taggedSeq) {
            System.out.println(token);
        }
        } finally {
            tagger.close();
        }

    }

}
