package de.upb.infexba.ml;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.TimeZone;

import org.apache.uima.UIMAFramework;
import org.apache.uima.analysis_engine.AnalysisEngine;
import org.apache.uima.cas.CAS;
import org.apache.uima.cas.FSIterator;
import org.apache.uima.cas.text.AnnotationIndex;
import org.apache.uima.collection.CollectionReader;
import org.apache.uima.jcas.JCas;
import org.apache.uima.jcas.tcas.Annotation;
import org.apache.uima.resource.ResourceSpecifier;
import org.apache.uima.util.XMLInputSource;

import de.upb.infexba.Sentence;
import de.upb.infexba.Token;
import de.upb.infexba.cr.XmiCollectionReader;
import de.upb.infexba.ml.feature.ngram.CharNgramHandler;
import de.upb.infexba.ml.feature.ngram.POSNgramHandler;
import de.upb.infexba.ml.feature.partofspeech.PartOfSpeechFeatureHandler;
import de.upb.infexba.ml.feature.vocabulary.CounterFeatureHandler;
import de.upb.infexba.ml.feature.vocabulary.VocabularyFeatureHandler;

/**
 * @author kabu
 * 
 */
public class GenerationOfNormalizationFactors {

	// -------------------------------------------------------------------
	// FIXED FILE PATHS AND NAMES
	// -------------------------------------------------------------------

	/**
	 * Location of collection reader
	 */
	private static String collectionReaderDesc = "descriptors/collection/RevenueCorpusReader.xml";

	/**
	 * Location of aggregate revenue statement extractor
	 */
	private static String analysisEngineDesc_fast = "descriptors/aggregate/AggregatePreprocessing.xml";

	private static String analysisEngineDesc_mate = "descriptors/aggregate/AggregateMatePreprocessing.xml";

	/**
	 * Path of input directory (split will be added in main)
	 */
	private static String inputDir = "corpora/lfa-corpus/";

	/**
	 * Path of the directory where the .arff files are stored
	 */
	private static String arffDir = "arff/";

	// -------------------------------------------------------------------
	// OPTIONS
	// -------------------------------------------------------------------

	/**
	 * file from that contains the features which maxims should be computed
	 */
	public static String arffFile = "dataset_intention_05-17 09 42 59_music-training.arff"; //"dataset_intention_05-17 10 17 40_lfa-corpus-corpus-smartphone_split-training.arff";

	/**
	 * which corpus?
	 */
	public static final boolean MUSIC = true, SMARTPHONE = false;
	public static boolean corpus = MUSIC;

	/**
	 * and from that corpus, which set? (should be MODE_TRAIN always... but you could others as well)
	 */
	public static final byte MODE_SAMPLE = -1, MODE_TRAIN = 0, MODE_VALIDATE = 1, MODE_TEST = 2;
	public static byte mode = GenerationOfNormalizationFactors.MODE_TRAIN;

	/**
	 * what kind of normalization factors shall be computed?
	 */
	private static final short FACTOR_MAX = 0, FACTOR_MITTELWERT = 1;

	private static short factor = FACTOR_MITTELWERT;

	/**
	 * Use mate?
	 */
	private static boolean use_mate = false;

	// -------------------------------------------------------------------
	// SETTINGS FOR MUSIC OR SMARTPHONE CORPUS
	// -------------------------------------------------------------------

	static final String SAMPLE = "music_split/sample";
	static String TRAINING = "";
	static String VALIDATION = "";
	static String TEST = "";
	static boolean GET_FIRST_ANNOTATION = false;
	static {
		if (corpus) {
			// music -> always use first annotation
			GET_FIRST_ANNOTATION = true;
			TRAINING = "music_split/training";
			VALIDATION = "music_split/validation";
			TEST = "music_split/test";
		} else {
			// smartphone -> always use second annotation
			GET_FIRST_ANNOTATION = false;
			TRAINING = "smartphone_split/training";
			VALIDATION = "smartphone_split/validation";
			TEST = "smartphone_split/test";
		}
	}

	// -------------------------------------------------------------------
	// MAIN METHOD – GENERATION METHOD
	// -------------------------------------------------------------------

	static PartOfSpeechFeatureHandler posFeatureHandler = new PartOfSpeechFeatureHandler();

	static VocabularyFeatureHandler vocFeatureHandler = new VocabularyFeatureHandler();

	static CounterFeatureHandler counterFeatureHandler = new CounterFeatureHandler();

	static POSNgramHandler posNgramFeatureHandler = new POSNgramHandler(3);

	static CharNgramHandler charNgramFeatureHandler = new CharNgramHandler(3);

	static HashMap<String, Double> maxMap = new HashMap<String, Double>();

	static HashMap<String, Double> sumMap = new HashMap<String, Double>();

	static List<String> featureNames = null;

	/**
	 * The main method executes the evaluation process. That is, it computes recall, precision and value that refer to these for the statements found using the
	 * extraction algorithm.
	 * <p>
	 * Additionally, the recall of the found sentence, time and money expressions is computed.
	 * <p>
	 * Both are printed to screen.
	 * 
	 * @param args Not used
	 */
	public static void main(String[] args) {

		String dataset = TRAINING;
		switch (mode) {
		case MODE_SAMPLE:
			dataset = SAMPLE;
			break;
		case MODE_TRAIN:
			dataset = TRAINING;
			break;
		case MODE_VALIDATE:
			dataset = VALIDATION;
			break;
		case MODE_TEST:
			dataset = TEST;
			break;
		}

		XMLInputSource xmlInputSource;
		ResourceSpecifier specifier;
		AnalysisEngine analysisEngine;
		try {
			// Create analysis engine
			String analysisEngineDesc = analysisEngineDesc_fast;
			if (use_mate)
				analysisEngineDesc = analysisEngineDesc_mate;
			xmlInputSource = new XMLInputSource(analysisEngineDesc);
			specifier = UIMAFramework.getXMLParser().parseResourceSpecifier(xmlInputSource);
			analysisEngine = UIMAFramework.produceAnalysisEngine(specifier);
			if (use_mate) {
				analysisEngine.setConfigParameterValue("MateToolsLemmatizerWrapper/Activated", "true");
				analysisEngine.setConfigParameterValue("MateToolsPOSTaggerWrapper/Activated", "true");
			}
			analysisEngine.setConfigParameterValue("RBTokenizer/Activated", "true");
			analysisEngine.setConfigParameterValue("RBSentenceSplitter/Activated", "true");
			analysisEngine.setConfigParameterValue("PartOfSpeechAndChunkTagger/Activated", "true");
			analysisEngine.setConfigParameterValue("TimeExpMatcher/Activated", "true");
			analysisEngine.setConfigParameterValue("MoneyExpMatcher/Activated", "true");
			analysisEngine.reconfigure();

			String curInputDir = inputDir + dataset + "/";

			// Initialization of FeatureHandlers
			initialize();

			// Get the collection to be iterated over
			xmlInputSource = new XMLInputSource(collectionReaderDesc);
			specifier = UIMAFramework.getXMLParser().parseResourceSpecifier(xmlInputSource);
			CollectionReader collectionReader = UIMAFramework.produceCollectionReader(specifier);
			collectionReader.setConfigParameterValue(XmiCollectionReader.PARAM_INPUTDIR, curInputDir);
			collectionReader.reconfigure();
			int numberOfDocuments = 0;
			if (dataset.equals(TRAINING) || dataset.equals(SAMPLE)) {
				System.out.print("Initialize feature set...");
				int doc = 1;
				CAS aCAS = analysisEngine.newCAS();
				int no_classification = 0;
				while (collectionReader.hasNext()) {
					doc++;
					if (doc % 10 == 0)
						System.out.print(".");
					collectionReader.getNext(aCAS);
					JCas jcas = aCAS.getJCas();
					// Process document
					analysisEngine.process(jcas);
					process(jcas);
					numberOfDocuments++;
				}
				System.out.print("skipped " + no_classification + " file(s) without classification ...");
				System.out.println("finished");
			}

			// get factors...
			String factors = getFactorString(factor, numberOfDocuments);

			// ...and write them to an outputFile
			SimpleDateFormat df = new SimpleDateFormat("MM-dd HH mm ss");
			df.setTimeZone(TimeZone.getDefault());
			String name = dataset.replace('/', '-');
			File protocol = new File(arffDir + "factors_" + df.format(new Date()) + "_" + name + ".txt");
			Writer out = new OutputStreamWriter(new FileOutputStream(protocol),"UTF-8");
			try {
				out.write("Faktoren basierend auf " + arffFile + "\n\n" + factors);
			} finally {
				out.close();
			}

			analysisEngine.destroy();

		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	private static void initialize() {
		// Read feature names from file
		GeneralFeatureARFFReader reader = new GeneralFeatureARFFReader();
		featureNames = reader.readFeatureSetFromARFF(arffDir + arffFile);
		// Initialize Handlers
		for (String feature : featureNames)
			if (posFeatureHandler.canHandle(feature))
				posFeatureHandler.addFeature(feature);
			else if (counterFeatureHandler.canHandle(feature))
				counterFeatureHandler.addFeature(feature);
			else if (vocFeatureHandler.canHandle(feature))
				vocFeatureHandler.addFeature(feature);
			else if (posNgramFeatureHandler.canHandle(feature))
				posNgramFeatureHandler.addFeature(feature);
			else if (charNgramFeatureHandler.canHandle(feature))
				charNgramFeatureHandler.addFeature(feature);
		// Initialize maxMap
		for (String feature : featureNames) {
			maxMap.put(feature, 0.);
			sumMap.put(feature, 0.);
		}
	}

	private static void process(JCas jcas) {
		// compute feature
		AnnotationIndex<Annotation> sentenceIndex = jcas.getAnnotationIndex(Sentence.type);
		AnnotationIndex<Annotation> tokenIndex = jcas.getAnnotationIndex(Token.type);
		FSIterator<Annotation> tokenIterator;
		FSIterator<Annotation> sentenceIterator = sentenceIndex.iterator();
		Sentence sentence;
		Token token;
		double number_of_tokens = 0.;
		while (sentenceIterator.hasNext()) {
			sentence = (Sentence) sentenceIterator.next();
			counterFeatureHandler.handle(sentence);
			tokenIterator = tokenIndex.subiterator(sentence);
			while (tokenIterator.hasNext()) {
				number_of_tokens++;
				token = (Token) tokenIterator.next();
				String word = token.getCoveredText();
				String lemma = token.getLemma();
				// part-of-speech
				String pos = PartOfSpeechFeatureHandler.getPos(token);
				posFeatureHandler.handle(pos);
				// word/lemma
				vocFeatureHandler.handle(word, lemma, pos, null);
				// counter per word
				counterFeatureHandler.handle(word);
			}
		}
		counterFeatureHandler.handle(jcas);
		counterFeatureHandler.computeFeatures();
		posNgramFeatureHandler.handle(jcas, null);
		charNgramFeatureHandler.handle(jcas, null);
		// update maxima
		updateMaxima(number_of_tokens);
		updateSums(number_of_tokens);
		// clear
		clearFeatureMapsOfAllHandlers();
	}

	private static void updateMaxima(double number_of_tokens) {
		for (String feature : featureNames) {
			if (posFeatureHandler.handles(feature))
				updateMaximum(feature, posFeatureHandler.getFromFeatureMap(feature) / number_of_tokens);
			else if (counterFeatureHandler.handles(feature))
				updateMaximum(feature, counterFeatureHandler.getFromFeatureMap(feature));
			else if (vocFeatureHandler.handles(feature))
				updateMaximum(feature, vocFeatureHandler.getFromFeatureMap(feature) / number_of_tokens);
			else if (posNgramFeatureHandler.handles(feature))
				updateMaximum(feature, (double) posNgramFeatureHandler.getFromFeatureMap(feature) / number_of_tokens);
			else if (charNgramFeatureHandler.handles(feature))
				updateMaximum(feature, (double) charNgramFeatureHandler.getFromFeatureMap(feature) / number_of_tokens);
		}
	}

	private static void updateMaximum(String feature, double value) {
		if (maxMap.containsKey(feature)) {
			if (maxMap.get(feature) < value)
				maxMap.put(feature, value);
		} else {
			System.err.println("Feature " + feature + " not initialized in maxMap.");
		}
	}

	private static void updateSums(double number_of_tokens) {
		for (String feature : featureNames) {
			if (posFeatureHandler.handles(feature))
				updateSum(feature, posFeatureHandler.getFromFeatureMap(feature) / number_of_tokens);
			else if (counterFeatureHandler.handles(feature))
				updateSum(feature, counterFeatureHandler.getFromFeatureMap(feature));
			else if (vocFeatureHandler.handles(feature))
				updateSum(feature, vocFeatureHandler.getFromFeatureMap(feature) / number_of_tokens);
			else if (posNgramFeatureHandler.handles(feature))
				updateSum(feature, (double) posNgramFeatureHandler.getFromFeatureMap(feature) / number_of_tokens);
			else if (charNgramFeatureHandler.handles(feature))
				updateSum(feature, (double) charNgramFeatureHandler.getFromFeatureMap(feature) / number_of_tokens);
		}
	}

	private static void updateSum(String feature, double value) {
		if (sumMap.containsKey(feature)) {
			sumMap.put(feature, value + sumMap.get(feature));
		} else {
			System.err.println("Feature " + feature + " not initialized in maxMap.");
		}
	}

	private static void clearFeatureMapsOfAllHandlers() {
		vocFeatureHandler.clearFeatureMap();
		posFeatureHandler.clearFeatureMap();
		counterFeatureHandler.clearFeatureMap();
		posNgramFeatureHandler.clearFeatureMap();
		charNgramFeatureHandler.clearFeatureMap();
	}

	private static String getFactorString(short kind, double numberOfDocuments) {
		String factors = "";
		if (kind == GenerationOfNormalizationFactors.FACTOR_MAX)
			for (String feature : featureNames)
				factors += "else if(feature.equals(\"" + feature + "\"))\n\tfactor = " + (1. / maxMap.get(feature)) + ";\n";
		else if (kind == GenerationOfNormalizationFactors.FACTOR_MITTELWERT)
			for (String feature : featureNames){
				double mittelwert = sumMap.get(feature) / numberOfDocuments;
				factors += "else if(feature.equals(\"" + feature + "\"))\n\tfactor = " + (0.5 / mittelwert) + ";\n";
			}
		return factors;

	}
}