package de.upb.infexba.ml.feature.vocabulary;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import de.upb.infexba.ml.IntentionFeatureCalculator;
import de.upb.infexba.ml.feature.FeatureHandler;
import de.upb.infexba.ml.feature.GenreFeatureHandler;
import de.upb.infexba.util.Tuple;

public class VocabularyFeatureHandler extends GenreFeatureHandler {

	public static final byte WORD_FEATURE = 0;

	public static final byte LEMMA_FEATURE = 1;

	private static final String WORD_FEATURE_SUFFIX = "|w";

	private static final String LEMMA_FEATURE_SUFFIX = "|l";

	private String STRING_SET_FEATURE_PREFIX = "STRING_SET_FEATURE_"; // for constructing feature names

	// -------------------------------------------------------------------
	// SPECIAL (VOCABULRAY) FEATURES
	// -------------------------------------------------------------------

	public static Set<String> XY_SYMBOLS;
	public static Set<String> QUOTE_LIST;
	public static Set<String> PARANTHESIS_LIST;
	static {
		// these symbols won't be recognized by the Tagger as "XY" that's why we filter them by hand
		String[] xy = { "…", "«", "" };
		XY_SYMBOLS = new HashSet<String>(Arrays.asList(xy));

		String[] quotes = { "„", "“", "”", "’", "'", "`", "´", "\"", ",\"", ".\"" };
		QUOTE_LIST = new HashSet<String>(Arrays.asList(quotes));

		String[] paranthesis = { ")", "(", "[", "]", "{", "}" };
		PARANTHESIS_LIST = new HashSet<String>(Arrays.asList(paranthesis));
	}

	// -------------------------------------------------------------------
	// MAPS
	// -------------------------------------------------------------------

	// information per class: lemma

	private HashMap<String, Double> commercial_lemma_map = new HashMap<String, Double>();
	private HashMap<String, Double> informational_lemma_map = new HashMap<String, Double>();
	private HashMap<String, Double> personal_lemma_map = new HashMap<String, Double>();

	// information per class: words

	private HashMap<String, Double> commercial_word_map = new HashMap<String, Double>();
	private HashMap<String, Double> informational_word_map = new HashMap<String, Double>();
	private HashMap<String, Double> personal_word_map = new HashMap<String, Double>();

	// information per class: information needed for IDF

	private HashSet<String> tmp_personal_word_set = new HashSet<String>();
	private HashSet<String> tmp_commercial_word_set = new HashSet<String>();
	private HashSet<String> tmp_informational_word_set = new HashSet<String>();
	private HashMap<String, Double> commercial_word_occures_in_doc_map = new HashMap<String, Double>();
	private HashMap<String, Double> informational_word_occures_in_doc_map = new HashMap<String, Double>();
	private HashMap<String, Double> personal_word_occures_in_doc_map = new HashMap<String, Double>();

	// # documents per genre

	private int number_of_commercial_documents = 0;
	private int number_of_informational_documents = 0;
	private int number_of_personal_documents = 0;

	// string set features

	public HashMap<String, HashSet<String>> string_set_features = new HashMap<String, HashSet<String>>();
	private int number_of_string_set_features = 0;

	// -------------------------------------------------------------------
	// Skip-Lists.
	// -------------------------------------------------------------------

	// tokens which word_key appears in the word_key_skip won't be taken into account!
	private Set<String> word_key_skip = new HashSet<String>();
	// tokens which lemma_key appears in the lemma_key_skip won't be taken into account!
	private Set<String> lemma_key_skip = new HashSet<String>();
	// tokens which pos appears in the pos_skip won't be taken into account!
	private Set<String> pos_skip = new HashSet<String>();

	public Set<String> getWordKeySkipList() {
		return word_key_skip;
	}

	public Set<String> getLemmaKeySkipList() {
		return lemma_key_skip;
	}

	public void setWordSkipList(String[] word_keys) {
		word_key_skip.clear();
		word_key_skip.addAll(Arrays.asList(word_keys));
	}

	public void setLemmaSkipList(String[] lemma_keys) {
		lemma_key_skip.clear();
		lemma_key_skip.addAll(Arrays.asList(lemma_keys));
	}

	public void setPOSSkipList(String[] pos_skip_list_for_tokens) {
		pos_skip.clear();
		pos_skip.addAll(Arrays.asList(pos_skip_list_for_tokens));
	}

	// -------------------------------------------------------------------
	// FeatureHandler.
	// -------------------------------------------------------------------

	public VocabularyFeatureHandler() {

	}

	public TreeSet<String> getFeatures() {
		System.err.println("VocabularyFeatureHandler: Please use get_top_words or get_genre_specific_words methods instead");
		return null;
	}

	public boolean canHandle(String feature) {
		return isWordOrLemma(feature);
	}

	// -------------------------------------------------------------------
	// KEY
	// -------------------------------------------------------------------

	/**
	 * Given a set of features names (i.e. words with suffix |w or |l) defines this set to be one feature and returns a name for this feature.
	 * 
	 * @param features
	 * @return
	 */
	public String define_as_one_feature(HashSet<String> feature_names) {
		number_of_string_set_features++;
		String new_feature_name = STRING_SET_FEATURE_PREFIX + number_of_string_set_features;
		string_set_features.put(new_feature_name, feature_names);
		return new_feature_name;
	}

	/**
	 * Returns the feature-name of a string feature. In order to distinguish between lemma and words we add a suffix to the feature (lemma/word) itself and the
	 * feature will be converted to lower-case. Additionally, if the word is contained in a so called feature-set then the name of the feature-set will be
	 * returned.
	 * 
	 * @param word a word or a word-key
	 * @param kind
	 * @return
	 */
	private String getFeatureName(String featureName, int kind) {
		String key = "";
		String text = getTextFromFeatureName(featureName);

		if (XY_SYMBOLS.contains(text) || PARANTHESIS_LIST.contains(text)) {
			text = "";
		} else if (text.equals("€")) {
			text = "euro";
		} else if (QUOTE_LIST.contains(text)) {
			text = "";
			// lemma_key = get_key(PUNCTATION_MARK, LEMMA_FEATURE);
			// word_key = get_key(QUOTE, WORD_FEATURE);
		}

		if (kind == WORD_FEATURE)
			key = text.toLowerCase() + WORD_FEATURE_SUFFIX;
		else if (kind == LEMMA_FEATURE)
			key = text.toLowerCase() + LEMMA_FEATURE_SUFFIX;
		else
			key = text.toLowerCase();

		// handling feature-sets
		if (!string_set_features.isEmpty())
			for (String string_set_feature : string_set_features.keySet())
				if (string_set_features.get(string_set_feature).contains(key))
					key = string_set_feature;

		// System.out.println("word="+word+" kind="+kind+" word-key"+key);

		return key;
	}

	public static String getTextFromFeatureName(String featureName) {
		if (featureName.indexOf('|') > -1)
			return featureName.substring(0, featureName.indexOf('|'));
		else
			return featureName.toLowerCase(); // !!!
	}

	public static boolean isWord(String feature) {
		return (feature.indexOf(WORD_FEATURE_SUFFIX) > -1);
	}

	public static boolean isLemma(String feature) {
		return (feature.indexOf(LEMMA_FEATURE_SUFFIX) > -1);
	}

	public static boolean isWordOrLemma(String feature) {
		return isWord(feature) || isLemma(feature);
	}

	
	public void handle(String word, String lemma, String pos, String classification){
		if (classification == null || classification.isEmpty()) {
			this.count(word,lemma,  pos);
		} else {
			// if we're working with the training-set (i.e. featureNames == null) then save
			// what we know about the words/lemmas
			this.training_count( word,lemma, pos, classification);
			this.training_notify_occurence(classification, word);
		}
	}
	
	
	// -------------------------------------------------------------------
	// COUNT OCCURENCE OF WORDS - during Classification
	// -------------------------------------------------------------------

	// ... use the feature_map

	public void count(String word, String lemma, String pos) {

		String lemma_key = getFeatureName(lemma, VocabularyFeatureHandler.LEMMA_FEATURE);
		String word_key = getFeatureName(word, VocabularyFeatureHandler.WORD_FEATURE);

		// first skip, if necessary
		if (!lemma_key_skip.contains(lemma_key) && !word_key_skip.contains(word_key) && !this.pos_skip.contains(pos)) { // &&
																														// word.length()
																														// !=
																														// 1)
																														// {
			this.increaseInFeatureMap(lemma_key, 1.);
			this.increaseInFeatureMap(word_key, 1.);
		}
	}

	// -------------------------------------------------------------------
	// COUNT OCCURENCE OF WORDS - during Training
	// -------------------------------------------------------------------

	// ... use the genre_word/lemma_map

	public void training_count(String word, String lemma, String pos, String classification) {

		String lemma_key = getFeatureName(lemma, VocabularyFeatureHandler.LEMMA_FEATURE);
		String word_key = getFeatureName(word, VocabularyFeatureHandler.WORD_FEATURE);

		if (!lemma_key_skip.contains(lemma_key) && !word_key_skip.contains(word_key) && !this.pos_skip.contains(pos)) {
			// && word.length()!=1){
			HashMap<String, Double> map = this.getMapByName(classification, LEMMA_FEATURE);
			if (map.containsKey(lemma_key))
				map.put(lemma_key, map.get(lemma_key) + 1.);
			else
				map.put(lemma_key, 1.);

			map = this.getMapByName(classification, WORD_FEATURE);
			if (map.containsKey(word_key))
				map.put(word_key, map.get(word_key) + 1.);
			else
				map.put(word_key, 1.);
		}
		// else{
		// System.out.println("skipped "+lemma+", "+word+", "+pos);
		// }
	}

	// count during training: per document

	public void training_notify_occurence(String classification, String word) {
		String word_key = getFeatureName(word, VocabularyFeatureHandler.WORD_FEATURE);
		if (classification.equals(IntentionFeatureCalculator.PERSONAL)) {
			if (!tmp_personal_word_set.contains(word_key))
				tmp_personal_word_set.add(word_key);
		} else if (classification.equals(IntentionFeatureCalculator.COMMERCIAL)) {
			if (!tmp_commercial_word_set.contains(word_key))
				tmp_commercial_word_set.add(word_key);
		} else if (classification.equals(IntentionFeatureCalculator.INFORMATIONAL)) {
			if (!tmp_informational_word_set.contains(word_key))
				tmp_informational_word_set.add(word_key);
		}
	}

	public void training_save_occurences() {
		this.save_occurence(IntentionFeatureCalculator.PERSONAL, tmp_personal_word_set);
		this.save_occurence(IntentionFeatureCalculator.COMMERCIAL, tmp_commercial_word_set);
		this.save_occurence(IntentionFeatureCalculator.INFORMATIONAL, tmp_informational_word_set);
		tmp_personal_word_set.clear();
		tmp_commercial_word_set.clear();
		tmp_informational_word_set.clear();
	}

	private void save_occurence(String classification, Set<String> words_that_occurred) {
		HashMap<String, Double> map = this.get_occurence_in_doc_map_by_name(classification);
		for (String key : words_that_occurred)
			if (map.containsKey(key))
				map.put(key, map.get(key) + 1);
			else
				map.put(key, 1.);
	}

	public void training_count_classification(String classification) {
		if (classification.equals(IntentionFeatureCalculator.COMMERCIAL))
			number_of_commercial_documents++;
		if (classification.equals(IntentionFeatureCalculator.INFORMATIONAL))
			number_of_informational_documents++;
		if (classification.equals(IntentionFeatureCalculator.PERSONAL))
			number_of_personal_documents++;
	}

	// -------------------------------------------------------------------
	// GENRE-SPECIFIC-WORDS -- IDF
	// following only works for words (not for lemma)
	// -------------------------------------------------------------------

	public TreeSet<String> get_all_genre_specific_words__idf(int T, double threshold) {
		TreeSet<String> genre_specific_words = new TreeSet<String>();
		genre_specific_words.addAll(get_genre_specific_words_for_one_class__idf(T, threshold, IntentionFeatureCalculator.PERSONAL,
				IntentionFeatureCalculator.COMMERCIAL, IntentionFeatureCalculator.INFORMATIONAL));
		genre_specific_words.addAll(get_genre_specific_words_for_one_class__idf(T, threshold, IntentionFeatureCalculator.COMMERCIAL,
				IntentionFeatureCalculator.PERSONAL, IntentionFeatureCalculator.INFORMATIONAL));
		genre_specific_words.addAll(get_genre_specific_words_for_one_class__idf(T, threshold, IntentionFeatureCalculator.INFORMATIONAL,
				IntentionFeatureCalculator.PERSONAL, IntentionFeatureCalculator.COMMERCIAL));

		return genre_specific_words;
	}

	// TODO: implement this as GenreFeatureHanlder method
	public List<String> get_genre_specific_words_for_one_class__idf(int T, double threshold, String name1, String name2, String name3) {

		HashMap<String, Double> map1 = get_occurence_in_doc_map_by_name(name1);
		HashMap<String, Double> map2 = get_occurence_in_doc_map_by_name(name2);
		HashMap<String, Double> map3 = get_occurence_in_doc_map_by_name(name3);

		int n1 = get_number_of_docs_by_name(name1);
		int n2 = get_number_of_docs_by_name(name1);
		int n3 = get_number_of_docs_by_name(name1);

		ArrayList<Tuple> tuple_list = new ArrayList<Tuple>();

		int others = n2 + n3;
		int others_containing_key = 0;
		for (String key : map1.keySet()) {
			if (map2.containsKey(key))
				others_containing_key += map2.get(key);
			if (map3.containsKey(key))
				others_containing_key += map3.get(key);
			tuple_list.add(new Tuple(key, map1.get(key) / n1 * Math.log(others / (1. + others_containing_key))));
		}
		Collections.sort(tuple_list);

		Iterator<Tuple> iter = tuple_list.iterator();
		ArrayList<String> list = new ArrayList<String>();
		int i = 0;
		while (iter.hasNext()) {
			Tuple t = iter.next();
			i++;

			if (((T > 0 && i <= T) || (T == 0)) && (t.getValue() > threshold))
				list.add(t.getString());
			else
				break;
		}

		return list;
	}

	// -------------------------------------------------------------------
	// HELPER-METHODS
	// -------------------------------------------------------------------

	private HashMap<String, Double> get_occurence_in_doc_map_by_name(String name) {
		if (name.equals(IntentionFeatureCalculator.PERSONAL))
			return this.personal_word_occures_in_doc_map;
		else if (name.equals(IntentionFeatureCalculator.COMMERCIAL))
			return this.commercial_word_occures_in_doc_map;
		else
			return this.informational_word_occures_in_doc_map;
	}

	private int get_number_of_docs_by_name(String name) {
		if (name.equals(IntentionFeatureCalculator.PERSONAL))
			return this.number_of_personal_documents;
		else if (name.equals(IntentionFeatureCalculator.COMMERCIAL))
			return this.number_of_commercial_documents;
		else
			return this.number_of_informational_documents;

	}

	protected HashMap<String, Double> getMapByName(String name, short kind) {
		if (kind == WORD_FEATURE) {
			if (name.equals(IntentionFeatureCalculator.PERSONAL))
				return personal_word_map;
			else if (name.equals(IntentionFeatureCalculator.COMMERCIAL))
				return commercial_word_map;
			else
				return informational_word_map;
		} else {
			if (name.equals(IntentionFeatureCalculator.PERSONAL))
				return personal_lemma_map;
			else if (name.equals(IntentionFeatureCalculator.COMMERCIAL))
				return commercial_lemma_map;
			else
				return informational_lemma_map;
		}
	}

	public double getNormalizedValue(String feature, double value, Double number_of_tokens) {
		if (this.normalizationMode == FeatureHandler.NO_FACTORS)
			return 1. * value / number_of_tokens;

		double factor = 1.;

		if (this.normalizationMode == FeatureHandler.MY_NORMALIZATION) {
			if (VocabularyFeatureHandler.isLemma(feature))
				factor = 10.; // lemma
			else
				factor = 100.; // word
		}else if(this.normalizationMode == FeatureHandler.CALCULATED_MAXIMA_MUSIC){
			if(feature.equals("beinahe|w|gs"))
				factor = 98.99999999999999;
			else if(feature.equals("boygroup|w|gs"))
				factor = 53.333333333333336;
			else if(feature.equals("charts|w|gs"))
				factor = 49.00000000000001;
			else if(feature.equals("dennoch|w|gs"))
				factor = 55.0;
			else if(feature.equals("derart|w|gs"))
				factor = 107.00000000000001;
			else if(feature.equals("einiges|w|gs"))
				factor = 80.0;
			else if(feature.equals("empfehlen|w|gs"))
				factor = 87.0;
			else if(feature.equals("fall|w|gs"))
				factor = 69.0;
			else if(feature.equals("fazit|w|gs"))
				factor = 110.0;
			else if(feature.equals("finde|w|gs"))
				factor = 68.0;
			else if(feature.equals("geeignet|w|gs"))
				factor = 85.0;
			else if(feature.equals("gekauft|w|gs"))
				factor = 101.0;
			else if(feature.equals("höre|w|gs"))
				factor = 30.0;
			else if(feature.equals("kaufen|w|gs"))
				factor = 87.0;
			else if(feature.equals("kauf|w|gs"))
				factor = 91.0;
			else if(feature.equals("klasse|w|gs"))
				factor = 30.0;
			else if(feature.equals("langsam|w|gs"))
				factor = 94.0;
			else if(feature.equals("legt|w|gs"))
				factor = 95.0;
			else if(feature.equals("lyrics|w|gs"))
				factor = 116.0;
			else if(feature.equals("muß|w|gs"))
				factor = 70.0;
			else if(feature.equals("passt|w|gs"))
				factor = 101.0;
			else if(feature.equals("review|w|gs"))
				factor = 30.0;
			else if(feature.equals("scheibe|w|gs"))
				factor = 45.142857142857146;
			else if(feature.equals("schlager|w|gs"))
				factor = 103.66666666666666;
			else if(feature.equals("studioalbum|w|gs"))
				factor = 95.0;
			else if(feature.equals("stücken|w|gs"))
				factor = 89.0;
			else if(feature.equals("vergleich|w|gs"))
				factor = 86.0;
			else if(feature.equals("wunder|w|gs"))
				factor = 67.5;
			else if(feature.equals("würden|w|gs"))
				factor = 91.0;
			else if(feature.equals("zumindest|w|gs"))
				factor = 63.0;
			else if(feature.equals("aber|w|top"))
				factor = 30.333333333333332;
			else if(feature.equals("album|w|top"))
				factor = 20.125;
			else if(feature.equals("als|w|top"))
				factor = 29.0;
			else if(feature.equals("an|w|top"))
				factor = 27.333333333333336;
			else if(feature.equals("auch|w|top"))
				factor = 27.400000000000002;
			else if(feature.equals("auf|w|top"))
				factor = 19.400000000000002;
			else if(feature.equals("aus|w|top"))
				factor = 32.0;
			else if(feature.equals("band|w|top"))
				factor = 34.0;
			else if(feature.equals("bei|w|top"))
				factor = 32.0;
			else if(feature.equals("cd|w|top"))
				factor = 29.666666666666668;
			else if(feature.equals("dass|w|top"))
				factor = 40.0;
			else if(feature.equals("das|w|top"))
				factor = 16.666666666666668;
			else if(feature.equals("dem|w|top"))
				factor = 23.4;
			else if(feature.equals("den|w|top"))
				factor = 24.0;
			else if(feature.equals("der|w|top"))
				factor = 12.833333333333334;
			else if(feature.equals("des|w|top"))
				factor = 25.666666666666668;
			else if(feature.equals("die|w|top"))
				factor = 14.875;
			else if(feature.equals("einem|w|top"))
				factor = 37.666666666666664;
			else if(feature.equals("einen|w|top"))
				factor = 36.33333333333333;
			else if(feature.equals("einer|w|top"))
				factor = 38.5;
			else if(feature.equals("eine|w|top"))
				factor = 21.0;
			else if(feature.equals("ein|w|top"))
				factor = 20.75;
			else if(feature.equals("er|w|top"))
				factor = 22.5;
			else if(feature.equals("es|w|top"))
				factor = 19.77777777777778;
			else if(feature.equals("für|w|top"))
				factor = 18.75;
			else if(feature.equals("hat|w|top"))
				factor = 31.5;
			else if(feature.equals("ich|w|top"))
				factor = 18.2;
			else if(feature.equals("im|w|top"))
				factor = 29.0;
			else if(feature.equals("in|w|top"))
				factor = 19.75;
			else if(feature.equals("ist|w|top"))
				factor = 23.799999999999997;
			else if(feature.equals("man|w|top"))
				factor = 24.57142857142857;
			else if(feature.equals("mir|w|top"))
				factor = 30.333333333333332;
			else if(feature.equals("mit|w|top"))
				factor = 14.5;
			else if(feature.equals("nach|w|top"))
				factor = 28.0;
			else if(feature.equals("nicht|w|top"))
				factor = 22.25;
			else if(feature.equals("noch|w|top"))
				factor = 34.0;
			else if(feature.equals("nur|w|top"))
				factor = 22.5;
			else if(feature.equals("oder|w|top"))
				factor = 32.75;
			else if(feature.equals("schon|w|top"))
				factor = 37.666666666666664;
			else if(feature.equals("sich|w|top"))
				factor = 27.666666666666668;
			else if(feature.equals("sie|w|top"))
				factor = 19.42857142857143;
			else if(feature.equals("sind|w|top"))
				factor = 31.666666666666664;
			else if(feature.equals("songs|w|top"))
				factor = 41.599999999999994;
			else if(feature.equals("so|w|top"))
				factor = 28.5;
			else if(feature.equals("und|w|top"))
				factor = 15.0;
			else if(feature.equals("von|w|top"))
				factor = 21.5;
			else if(feature.equals("wie|w|top"))
				factor = 35.666666666666664;
			else if(feature.equals("zu|w|top"))
				factor = 22.714285714285715;
		}else if(this.normalizationMode == FeatureHandler.CALCULATED_MAXIMA_SMARTPHONE){
			if(feature.equals(".com|w|gs"))
				factor = 28.0;
			else if(feature.equals("august|w|gs"))
				factor = 32.0;
			else if(feature.equals("bin|w|gs"))
				factor = 34.0;
			else if(feature.equals("denn|w|gs"))
				factor = 30.0;
			else if(feature.equals("download|w|gs"))
				factor = 39.0;
			else if(feature.equals("eigene|w|gs"))
				factor = 50.0;
			else if(feature.equals("entwickler|w|gs"))
				factor = 37.0;
			else if(feature.equals("er|w|gs"))
				factor = 25.5;
			else if(feature.equals("finde|w|gs"))
				factor = 45.0;
			else if(feature.equals("firmware|w|gs"))
				factor = 18.0;
			else if(feature.equals("gesehen|w|gs"))
				factor = 110.0;
			else if(feature.equals("hab|w|gs"))
				factor = 54.0;
			else if(feature.equals("ich|w|gs"))
				factor = 13.5;
			else if(feature.equals("ihre|w|gs"))
				factor = 42.0;
			else if(feature.equals("iphoneblog|w|gs"))
				factor = 52.0;
			else if(feature.equals("irgendwie|w|gs"))
				factor = 63.0;
			else if(feature.equals("jahr|w|gs"))
				factor = 13.0;
			else if(feature.equals("kÃ¶nnte|w|gs"))
				factor = 21.0;
			else if(feature.equals("machen|w|gs"))
				factor = 25.0;
			else if(feature.equals("mag|w|gs"))
				factor = 59.0;
			else if(feature.equals("meinem|w|gs"))
				factor = 65.0;
			else if(feature.equals("mich|w|gs"))
				factor = 31.0;
			else if(feature.equals("mir|w|gs"))
				factor = 25.0;
			else if(feature.equals("notizen|w|gs"))
				factor = 27.0;
			else if(feature.equals("nun|w|gs"))
				factor = 27.0;
			else if(feature.equals("os|w|gs"))
				factor = 20.0;
			else if(feature.equals("sogar|w|gs"))
				factor = 51.0;
			else if(feature.equals("sollen|w|gs"))
				factor = 26.0;
			else if(feature.equals("springer|w|gs"))
				factor = 136.72727272727272;
			else if(feature.equals("tim|w|gs"))
				factor = 95.0;
			else if(feature.equals("verlag|w|gs"))
				factor = 241.4;
			else if(feature.equals("verÃ¶ffentlicht|w|gs"))
				factor = 17.0;
			else if(feature.equals("viele|w|gs"))
				factor = 28.0;
			else if(feature.equals("welt|w|gs"))
				factor = 37.0;
			else if(feature.equals("wordpress|w|gs"))
				factor = 18.42857142857143;
			else if(feature.equals("zeitung|w|gs"))
				factor = 70.0;
			else if(feature.equals("Â |w|gs"))
				factor = 16.62962962962963;
			else if(feature.equals(".de|w|top"))
				factor = 23.333333333333332;
			else if(feature.equals("aber|w|top"))
				factor = 26.5;
			else if(feature.equals("ab|w|top"))
				factor = 24.500000000000004;
			else if(feature.equals("als|w|top"))
				factor = 16.0;
			else if(feature.equals("an|w|top"))
				factor = 18.0;
			else if(feature.equals("auch|w|top"))
				factor = 19.5;
			else if(feature.equals("auf|w|top"))
				factor = 14.5;
			else if(feature.equals("aus|w|top"))
				factor = 16.5;
			else if(feature.equals("bei|w|top"))
				factor = 21.5;
			else if(feature.equals("dass|w|top"))
				factor = 29.5;
			else if(feature.equals("das|w|top"))
				factor = 9.5;
			else if(feature.equals("dem|w|top"))
				factor = 16.5;
			else if(feature.equals("den|w|top"))
				factor = 8.333333333333334;
			else if(feature.equals("der|w|top"))
				factor = 11.333333333333332;
			else if(feature.equals("des|w|top"))
				factor = 16.0;
			else if(feature.equals("die|w|top"))
				factor = 11.5;
			else if(feature.equals("einem|w|top"))
				factor = 30.5;
			else if(feature.equals("einen|w|top"))
				factor = 19.5;
			else if(feature.equals("eine|w|top"))
				factor = 17.0;
			else if(feature.equals("ein|w|top"))
				factor = 22.666666666666664;
			else if(feature.equals("es|w|top"))
				factor = 10.0;
			else if(feature.equals("euro|w|top"))
				factor = 21.666666666666664;
			else if(feature.equals("für|w|top"))
				factor = 12.0;
			else if(feature.equals("google|w|top"))
				factor = 10.333333333333334;
			else if(feature.equals("hat|w|top"))
				factor = 14.5;
			else if(feature.equals("im|w|top"))
				factor = 17.0;
			else if(feature.equals("in|w|top"))
				factor = 10.333333333333334;
			else if(feature.equals("iphone|w|top"))
				factor = 8.5;
			else if(feature.equals("ist|w|top"))
				factor = 17.0;
			else if(feature.equals("kann|w|top"))
				factor = 23.5;
			else if(feature.equals("man|w|top"))
				factor = 23.0;
			else if(feature.equals("mit|w|top"))
				factor = 10.0;
			else if(feature.equals("neue|w|top"))
				factor = 14.0;
			else if(feature.equals("nicht|w|top"))
				factor = 20.5;
			else if(feature.equals("noch|w|top"))
				factor = 13.0;
			else if(feature.equals("nur|w|top"))
				factor = 27.0;
			else if(feature.equals("oder|w|top"))
				factor = 27.0;
			else if(feature.equals("sich|w|top"))
				factor = 20.666666666666668;
			else if(feature.equals("sind|w|top"))
				factor = 33.0;
			else if(feature.equals("so|w|top"))
				factor = 34.0;
			else if(feature.equals("t-mobile|w|top"))
				factor = 13.0;
			else if(feature.equals("um|w|top"))
				factor = 22.5;
			else if(feature.equals("und|w|top"))
				factor = 8.6;
			else if(feature.equals("von|w|top"))
				factor = 17.666666666666668;
			else if(feature.equals("was|w|top"))
				factor = 32.5;
			else if(feature.equals("werden|w|top"))
				factor = 28.5;
			else if(feature.equals("wie|w|top"))
				factor = 25.666666666666668;
			else if(feature.equals("wird|w|top"))
				factor = 20.0;
			else if(feature.equals("wir|w|top"))
				factor = 27.666666666666668;
			else if(feature.equals("zum|w|top"))
				factor = 24.0;
			else if(feature.equals("zu|w|top"))
				factor = 11.0;
			else if(feature.equals("Ã¼ber|w|top"))
				factor = 18.0;

		}else if (this.normalizationMode == FeatureHandler.CALCULATED_AVERAGE_MUSIC){
			 if(feature.equals("beinahe|w|gs"))
					factor = 4409.462127677448;
				else if(feature.equals("boygroup|w|gs"))
					factor = 6341.267884854123;
				else if(feature.equals("charts|w|gs"))
					factor = 2307.697347187512;
				else if(feature.equals("dennoch|w|gs"))
					factor = 1784.334140372602;
				else if(feature.equals("derart|w|gs"))
					factor = 4813.177903628897;
				else if(feature.equals("einiges|w|gs"))
					factor = 4423.283562880828;
				else if(feature.equals("empfehlen|w|gs"))
					factor = 4100.023386553856;
				else if(feature.equals("fall|w|gs"))
					factor = 1329.1872760452409;
				else if(feature.equals("fazit|w|gs"))
					factor = 4524.689611845718;
				else if(feature.equals("finde|w|gs"))
					factor = 2678.626428953159;
				else if(feature.equals("geeignet|w|gs"))
					factor = 6703.918697221371;
				else if(feature.equals("gekauft|w|gs"))
					factor = 5024.568664812325;
				else if(feature.equals("höre|w|gs"))
					factor = 3094.54579977672;
				else if(feature.equals("kaufen|w|gs"))
					factor = 2763.1655720638755;
				else if(feature.equals("kauf|w|gs"))
					factor = 6308.162213412866;
				else if(feature.equals("klasse|w|gs"))
					factor = 2211.4014413073064;
				else if(feature.equals("langsam|w|gs"))
					factor = 4766.823889035609;
				else if(feature.equals("legt|w|gs"))
					factor = 3509.9729282240373;
				else if(feature.equals("lyrics|w|gs"))
					factor = 4364.880538452614;
				else if(feature.equals("muß|w|gs"))
					factor = 4123.713830131536;
				else if(feature.equals("passt|w|gs"))
					factor = 3210.5181628789683;
				else if(feature.equals("review|w|gs"))
					factor = 236.16584435234395;
				else if(feature.equals("scheibe|w|gs"))
					factor = 823.064823950703;
				else if(feature.equals("schlager|w|gs"))
					factor = 10062.379403658862;
				else if(feature.equals("studioalbum|w|gs"))
					factor = 5537.728065036444;
				else if(feature.equals("stücken|w|gs"))
					factor = 2890.4330383069114;
				else if(feature.equals("vergleich|w|gs"))
					factor = 3367.1890820301114;
				else if(feature.equals("wunder|w|gs"))
					factor = 5117.283843782413;
				else if(feature.equals("würden|w|gs"))
					factor = 4058.3733751028662;
				else if(feature.equals("zumindest|w|gs"))
					factor = 2052.848965238875;
				else if(feature.equals("aber|w|top"))
					factor = 146.81544298463194;
				else if(feature.equals("album|w|top"))
					factor = 118.6615276710528;
				else if(feature.equals("als|w|top"))
					factor = 129.49029029733168;
				else if(feature.equals("an|w|top"))
					factor = 137.33193052259813;
				else if(feature.equals("auch|w|top"))
					factor = 85.07628976620867;
				else if(feature.equals("auf|w|top"))
					factor = 65.81423205595833;
				else if(feature.equals("aus|w|top"))
					factor = 152.02013209866905;
				else if(feature.equals("band|w|top"))
					factor = 292.4546333484148;
				else if(feature.equals("bei|w|top"))
					factor = 181.738480169502;
				else if(feature.equals("cd|w|top"))
					factor = 273.43938619956776;
				else if(feature.equals("dass|w|top"))
					factor = 191.72063854117536;
				else if(feature.equals("das|w|top"))
					factor = 49.15219339719;
				else if(feature.equals("dem|w|top"))
					factor = 104.57602846361141;
				else if(feature.equals("den|w|top"))
					factor = 60.74462928809594;
				else if(feature.equals("der|w|top"))
					factor = 28.770415925895612;
				else if(feature.equals("des|w|top"))
					factor = 178.79450043090915;
				else if(feature.equals("die|w|top"))
					factor = 23.364411927493233;
				else if(feature.equals("einem|w|top"))
					factor = 246.1633346132942;
				else if(feature.equals("einen|w|top"))
					factor = 241.29942674564896;
				else if(feature.equals("einer|w|top"))
					factor = 281.7297483268324;
				else if(feature.equals("eine|w|top"))
					factor = 109.37090623776919;
				else if(feature.equals("ein|w|top"))
					factor = 78.34052995330758;
				else if(feature.equals("er|w|top"))
					factor = 222.2636976521353;
				else if(feature.equals("es|w|top"))
					factor = 88.5108312597015;
				else if(feature.equals("für|w|top"))
					factor = 116.02366969386003;
				else if(feature.equals("hat|w|top"))
					factor = 151.64717468243637;
				else if(feature.equals("ich|w|top"))
					factor = 109.99228541115592;
				else if(feature.equals("im|w|top"))
					factor = 124.87816773696562;
				else if(feature.equals("in|w|top"))
					factor = 52.929130161609784;
				else if(feature.equals("ist|w|top"))
					factor = 59.503612938243656;
				else if(feature.equals("man|w|top"))
					factor = 113.93179572153353;
				else if(feature.equals("mir|w|top"))
					factor = 359.950233003711;
				else if(feature.equals("mit|w|top"))
					factor = 51.286792074739544;
				else if(feature.equals("nach|w|top"))
					factor = 203.50306691275415;
				else if(feature.equals("nicht|w|top"))
					factor = 78.02437851626254;
				else if(feature.equals("noch|w|top"))
					factor = 152.0289407763322;
				else if(feature.equals("nur|w|top"))
					factor = 189.0711719184201;
				else if(feature.equals("oder|w|top"))
					factor = 211.69430582700068;
				else if(feature.equals("schon|w|top"))
					factor = 181.99884071619866;
				else if(feature.equals("sich|w|top"))
					factor = 72.48708275007458;
				else if(feature.equals("sie|w|top"))
					factor = 151.2210941112619;
				else if(feature.equals("sind|w|top"))
					factor = 173.22143528449593;
				else if(feature.equals("songs|w|top"))
					factor = 245.27983255479842;
				else if(feature.equals("so|w|top"))
					factor = 133.99626043611966;
				else if(feature.equals("und|w|top"))
					factor = 24.059959663476146;
				else if(feature.equals("von|w|top"))
					factor = 76.05874171078196;
				else if(feature.equals("wie|w|top"))
					factor = 108.05430394503846;
				else if(feature.equals("zu|w|top"))
					factor = 67.47991919896205;
				else 
					System.err.println("Fehler bei Voc "+feature);

		}else if (this.normalizationMode == FeatureHandler.CALCULATED_AVERAGE_SMARTPHONE){
			if(feature.equals(".com|w|gs"))
				factor = 1115.9202525852604;
			else if(feature.equals("august|w|gs"))
				factor = 1033.3033111896273;
			else if(feature.equals("bin|w|gs"))
				factor = 825.8362418516066;
			else if(feature.equals("denn|w|gs"))
				factor = 622.8281520378007;
			else if(feature.equals("download|w|gs"))
				factor = 920.9715203136911;
			else if(feature.equals("eigene|w|gs"))
				factor = 1802.3811657346146;
			else if(feature.equals("entwickler|w|gs"))
				factor = 940.31807751655;
			else if(feature.equals("er|w|gs"))
				factor = 662.0968760785556;
			else if(feature.equals("finde|w|gs"))
				factor = 2915.0986646594542;
			else if(feature.equals("firmware|w|gs"))
				factor = 529.420777960049;
			else if(feature.equals("gesehen|w|gs"))
				factor = 5352.928517604267;
			else if(feature.equals("hab|w|gs"))
				factor = 1935.990064856341;
			else if(feature.equals("ich|w|gs"))
				factor = 94.30425506527025;
			else if(feature.equals("ihre|w|gs"))
				factor = 1927.2145090716276;
			else if(feature.equals("iphoneblog|w|gs"))
				factor = 3141.1238962749467;
			else if(feature.equals("irgendwie|w|gs"))
				factor = 3680.5436599187606;
			else if(feature.equals("jahr|w|gs"))
				factor = 809.1510895333753;
			else if(feature.equals("könnte|w|gs"))
				factor = 897.5154573774257;
			else if(feature.equals("machen|w|gs"))
				factor = 945.4251012296652;
			else if(feature.equals("mag|w|gs"))
				factor = 3747.357120318598;
			else if(feature.equals("meinem|w|gs"))
				factor = 2189.7908882024767;
			else if(feature.equals("mich|w|gs"))
				factor = 671.144500683782;
			else if(feature.equals("mir|w|gs"))
				factor = 470.36181483497967;
			else if(feature.equals("notizen|w|gs"))
				factor = 1519.3853607565088;
			else if(feature.equals("nun|w|gs"))
				factor = 284.1165637498594;
			else if(feature.equals("os|w|gs"))
				factor = 578.7073496093461;
			else if(feature.equals("sogar|w|gs"))
				factor = 1706.490228616959;
			else if(feature.equals("sollen|w|gs"))
				factor = 1115.7029269297495;
			else if(feature.equals("springer|w|gs"))
				factor = 36446.03396282958;
			else if(feature.equals("tim|w|gs"))
				factor = 15172.953573165949;
			else if(feature.equals("verlag|w|gs"))
				factor = 22587.267462105894;
			else if(feature.equals("veröffentlicht|w|gs"))
				factor = 917.2879160596727;
			else if(feature.equals("viele|w|gs"))
				factor = 1055.9607276621314;
			else if(feature.equals("welt|w|gs"))
				factor = 1616.7430380974065;
			else if(feature.equals("wordpress|w|gs"))
				factor = 1969.1012538020875;
			else if(feature.equals("zeitung|w|gs"))
				factor = 10897.056786211546;
			else if(feature.equals(" |w|gs"))
				factor = 1102.159578156743;
			else if(feature.equals(".de|w|top"))
				factor = 484.66471022939584;
			else if(feature.equals("aber|w|top"))
				factor = 193.66486792373462;
			else if(feature.equals("ab|w|top"))
				factor = 538.8226700268158;
			else if(feature.equals("als|w|top"))
				factor = 179.5672457917766;
			else if(feature.equals("an|w|top"))
				factor = 181.56095511576842;
			else if(feature.equals("auch|w|top"))
				factor = 93.75212850517043;
			else if(feature.equals("auf|w|top"))
				factor = 60.35288718904761;
			else if(feature.equals("aus|w|top"))
				factor = 211.3674263827435;
			else if(feature.equals("bei|w|top"))
				factor = 145.5534736712199;
			else if(feature.equals("dass|w|top"))
				factor = 201.41359669021688;
			else if(feature.equals("das|w|top"))
				factor = 30.657084867606187;
			else if(feature.equals("dem|w|top"))
				factor = 75.73044889226412;
			else if(feature.equals("den|w|top"))
				factor = 60.491928922467565;
			else if(feature.equals("der|w|top"))
				factor = 32.53401736734414;
			else if(feature.equals("des|w|top"))
				factor = 117.10150259720969;
			else if(feature.equals("die|w|top"))
				factor = 25.802717240101014;
			else if(feature.equals("einem|w|top"))
				factor = 258.5146551653127;
			else if(feature.equals("einen|w|top"))
				factor = 182.86711336032405;
			else if(feature.equals("eine|w|top"))
				factor = 98.26202823451302;
			else if(feature.equals("ein|w|top"))
				factor = 70.05467968914547;
			else if(feature.equals("es|w|top"))
				factor = 80.69017565475916;
			else if(feature.equals("euro|w|top"))
				factor = 398.96417130654095;
			else if(feature.equals("für|w|top"))
				factor = 51.780020602908515;
			else if(feature.equals("google|w|top"))
				factor = 168.17926777593232;
			else if(feature.equals("hat|w|top"))
				factor = 126.69112535915271;
			else if(feature.equals("im|w|top"))
				factor = 91.83263248884789;
			else if(feature.equals("in|w|top"))
				factor = 50.26226965077978;
			else if(feature.equals("iphone|w|top"))
				factor = 26.617304308017058;
			else if(feature.equals("ist|w|top"))
				factor = 64.40944842654571;
			else if(feature.equals("kann|w|top"))
				factor = 198.03519507042202;
			else if(feature.equals("man|w|top"))
				factor = 138.90442946466962;
			else if(feature.equals("mit|w|top"))
				factor = 50.70904613131642;
			else if(feature.equals("neue|w|top"))
				factor = 249.3165266176269;
			else if(feature.equals("nicht|w|top"))
				factor = 88.29414802976459;
			else if(feature.equals("noch|w|top"))
				factor = 148.97186891813155;
			else if(feature.equals("nur|w|top"))
				factor = 217.95264482969705;
			else if(feature.equals("oder|w|top"))
				factor = 221.6822259623087;
			else if(feature.equals("sich|w|top"))
				factor = 98.56051536615131;
			else if(feature.equals("sind|w|top"))
				factor = 278.8763672688323;
			else if(feature.equals("so|w|top"))
				factor = 187.6990712077097;
			else if(feature.equals("t-mobile|w|top"))
				factor = 189.55805936683063;
			else if(feature.equals("um|w|top"))
				factor = 239.2993996836216;
			else if(feature.equals("und|w|top"))
				factor = 29.112760049424335;
			else if(feature.equals("von|w|top"))
				factor = 70.86029697618348;
			else if(feature.equals("was|w|top"))
				factor = 308.2314609091119;
			else if(feature.equals("werden|w|top"))
				factor = 203.32262414650256;
			else if(feature.equals("wie|w|top"))
				factor = 166.8912516811248;
			else if(feature.equals("wird|w|top"))
				factor = 145.3289853384341;
			else if(feature.equals("wir|w|top"))
				factor = 378.65759807231376;
			else if(feature.equals("zum|w|top"))
				factor = 187.02427688205336;
			else if(feature.equals("zu|w|top"))
				factor = 81.24785262478022;
			else if(feature.equals("über|w|top"))
				factor = 173.73918504293218;
			else 
				System.err.println("Fehler bei Voc "+feature);

		}

		return factor * value / number_of_tokens;

	}
}
