package de.upb.infexba.util;

/**
 * <p>
 * Helper class that offers several functions on characters,
 * mainly for determining if a given character belongs to a
 * certain category of characters.
 * 
 * @author hwachsmuth
 *
 */
public class CharTools {

	/**
	 * Determines whether <code>c</code> is a period, a
	 * question mark or an exclamation mark.
	 * 
	 * @param c The character
	 * @return <code>true</code> iff. <code>c</code> is a
	 * full stop.
	 */
	public static boolean isFullStop(char c){
		return (isPeriod(c) || isUmambiguousEndMark(c));
	}	

	/**
	 * Determines whether <code>c</code> is one of the
	 * characters, which can be seen as unambiguous end
	 * mark, i.e. it is a question mark or an exclamation 
	 * mark.
	 * 
	 * @param c The character
	 * @return <code>true</code> iff. <code>c</code> is an unambiguous end mark
	 */
	public static boolean isUmambiguousEndMark(char c){
		return (c == '?' || c == '!');
	}
	
	/**
	 * Determines whether <code>c</code> is a period.
	 * 
	 * @param c The character
	 * @return <code>true</code> iff. <code>c</code> is a period.
	 */
	public static boolean isPeriod(char c){
		return c == '.';
	}
	
	/**
	 * Determines whether <code>c</code> is a comma.
	 * 
	 * @param c The character
	 * @return <code>true</code> iff. <code>c</code> is a comma.
	 */
	public static boolean isComma(char c){
		return c == ',';
	}
	
	/**
	 * Determines whether <code>c</code> is one of the
	 * versions of hyphens.
	 * 
	 * @param c The character
	 * @return <code>true</code> iff. <code>c</code> is a hyphen
	 */
	public static boolean isHyphen(char c){
		return c == '-' || c == '–' || c == '—';
	}
	
	/**
	 * Determines whether <code>c</code> is any of the
	 * existing quotation mark characters.
	 * 
	 * @param c The character
	 * @return <code>true</code> iff. <code>c</code> is a quaotation mark.
	 */
	public static boolean isQuotationMark(char c){
		return c == '"' || c == '\'' || c == '“' || c == '„';
	}
	
	/**
	 * Determines whether <code>c</code> is one of the
	 * four types of an opening bracket.
	 * 
	 * @param c The character
	 * @return <code>true</code> iff. <code>c</code> is an opening bracket.
	 */
	public static boolean isOpeningBracket(char c){
		return c == '(' || c == '[' || c == '{' || c == '<';
	}

	/**
	 * Determines whether <code>c</code> is one of the
	 * four types of a closing bracket.
	 * 
	 * @param c The character
	 * @return <code>true</code> iff. <code>c</code> is a closing bracket.
	 */
	public static boolean isClosingBracket(char c){
		return c == ')' || c == ']' || c == '}' || c == '>';
	}
	
	/**
	 * Determines whether <code>c</code> is possible icon
	 * of a bulletpoint of an unordererd list.
	 * 
	 * @param c The character
	 * @return <code>true</code> iff. <code>c</code> is a bullet.
	 */
	public static boolean isBullet(char c){
		return c == '-' || c == '*' || c == '+' || c == '•' || c == '–';
	}
	
	/**
	 * Determines whether <code>c</code> is tendency symbol,
	 * i.e. plus or a minus.
	 * 
	 * @param c The character
	 * @return <code>true</code> iff. <code>c</code> is a tendency symbol
	 */
	public static boolean isTendencySymbol(char c){
		return c == '-' || c == '+';
	}
	
	/**
	 * Determines whether <code>c</code> is one of the most common
	 * currency symbols, namely €, $, £ or ¥.
	 * 
	 * @param c The character
	 * @return <code>true</code> iff. <code>c</code> is €, $, £ or ¥
	 */
	public static boolean isCurrencySymbol(char c){
		return c == '€' || c == '$' || c  == '£' || c == '¥'; 
	}
	
	/**
	 * Determines whether <code>c</code> is an "@".
	 * 
	 * @param c The character
	 * @return <code>true</code> iff. <code>c</code> is "@".
	 */
	public static boolean isAt(char c){
		return c == '@';
	}
	
	/**
	 * Determines whether <code>c</code> is the beginning
	 * of a comment
	 * 
	 * @param c The character
	 * @return <code>true</code> iff. <code>c</code> is a comment beginning.
	 */
	public static boolean isComment(char c){
		return c == '#';
	}
}
